from __future__ import annotations

from collections.abc import Iterable, Sequence
from contextlib import contextmanager
from contextvars import ContextVar
from dataclasses import dataclass
from enum import Enum
from threading import Lock
from typing import TYPE_CHECKING, Any, Literal, Self

from sonolus.backend.blocks import BlockData, PlayBlock
from sonolus.backend.ir import IRConst, IRExpr, IRStmt
from sonolus.backend.mode import Mode
from sonolus.backend.optimize.flow import BasicBlock, FlowEdge
from sonolus.backend.place import (
    PREALLOCATED_TEMP_PLACE_COUNT,
    Block,
    BlockPlace,
    TempBlock,
    preallocated_temp_block_places,
)
from sonolus.script.globals import _GlobalInfo, _GlobalPlaceholder
from sonolus.script.internal.value import Value

if TYPE_CHECKING:
    from sonolus.script.project import BuildConfig

_compiler_internal_ = True

context_var: ContextVar[Context | None] = ContextVar("context_var", default=None)  # type: ignore


@dataclass(frozen=True)
class DebugConfig:
    unchecked_reads: bool
    unchecked_writes: bool


_full_debug_config = DebugConfig(
    unchecked_reads=True,
    unchecked_writes=True,
)

_disabled_debug_config = DebugConfig(
    unchecked_reads=False,
    unchecked_writes=False,
)

debug_var = ContextVar("debug_var", default=_disabled_debug_config)


class RuntimeChecks(Enum):
    """Runtime error checking modes."""

    NONE = "none"
    """No runtime checks."""

    TERMINATE = "terminate"
    """Terminate on errors."""

    NOTIFY_AND_TERMINATE = "notify_and_terminate"
    """Log, debug pause, and terminate on errors."""


@dataclass
class FunctionVisitStatistics:
    total_time: int = 0
    own_time: int = 0
    call_count: int = 0


class ProjectContextState:
    rom: ReadOnlyMemory
    const_mappings: dict[Any, int]
    debug_str_mappings: dict[str, int]
    lock: Lock
    runtime_checks: RuntimeChecks
    visit_stats: dict[str, FunctionVisitStatistics]

    def __init__(
        self,
        rom: ReadOnlyMemory | None = None,
        const_mappings: dict[Any, int] | None = None,
        debug_str_mappings: dict[str, int] | None = None,
        runtime_checks: RuntimeChecks = RuntimeChecks.NONE,
    ):
        self.rom = ReadOnlyMemory() if rom is None else rom
        self.const_mappings = {} if const_mappings is None else const_mappings
        self.debug_str_mappings = {} if debug_str_mappings is None else debug_str_mappings
        self.lock = Lock()
        self.runtime_checks = runtime_checks
        self.visit_stats = {}

    @classmethod
    def from_build_config(
        cls,
        config: BuildConfig,
        rom: ReadOnlyMemory | None = None,
        const_mappings: dict[Any, int] | None = None,
        debug_str_mappings: dict[str, int] | None = None,
    ) -> Self:
        return cls(
            rom=rom,
            const_mappings=const_mappings,
            debug_str_mappings=debug_str_mappings,
            runtime_checks=config.runtime_checks,
        )


class ModeContextState:
    archetypes: dict[type, int]
    compile_time_only_archetypes: set[type]
    archetypes_by_name: dict[str, type]
    keys_by_archetype_id: Sequence[int]
    is_scored_by_archetype_id: Sequence[bool]
    archetype_mro_id_array_rom_indexes: Sequence[int] | None = None
    environment_mappings: dict[_GlobalInfo, int]
    environment_offsets: dict[Block, int]
    mode: Mode
    lock: Lock

    def __init__(self, mode: Mode, archetypes: list[type] | None = None):
        from sonolus.script.array import Array

        archetypes = [*archetypes] if archetypes is not None else []
        seen_archetypes = {*archetypes}
        compile_time_only_archetypes = set()
        for type_ in [*archetypes]:
            for entry in type_.mro():
                if getattr(entry, "_is_concrete_archetype_", False) and entry not in seen_archetypes:
                    archetypes.append(entry)
                    seen_archetypes.add(entry)
                    compile_time_only_archetypes.add(entry)
        self.archetypes = {type_: idx for idx, type_ in enumerate(archetypes)}
        self.compile_time_only_archetypes = compile_time_only_archetypes
        self.archetypes_by_name = {type_.name: type_ for type_, _ in self.archetypes.items()}  # type: ignore
        self.keys_by_archetype_id = (
            Array(*((getattr(a, "_key_", -1)) for a in archetypes)) if archetypes else Array[int, Literal[0]]()
        )
        self.is_scored_by_archetype_id = (
            Array(*((getattr(a, "_is_scored_", False)) for a in archetypes))
            if archetypes
            else Array[bool, Literal[0]]()
        )
        self.environment_mappings = {}
        self.environment_offsets = {}
        self.mode = mode
        self.lock = Lock()

    def _init_archetype_mro_info(self, rom: ReadOnlyMemory):
        from sonolus.script.array import Array
        from sonolus.script.num import Num

        with self.lock:
            if self.archetype_mro_id_array_rom_indexes is not None:
                return
            archetype_mro_id_values = []
            archetype_mro_id_offsets = []
            for type_ in self.archetypes:
                mro_ids = [self.archetypes[entry] for entry in type_.mro() if entry in self.archetypes]
                archetype_mro_id_offsets.append(len(archetype_mro_id_values))
                archetype_mro_id_values.append(len(mro_ids))
                archetype_mro_id_values.extend(mro_ids)
            archetype_mro_id_array_place = rom[tuple(archetype_mro_id_values)]

            archetype_mro_id_rom_indexes = Array[int, len(archetype_mro_id_offsets)]._with_value(
                [Num._accept_(offset + archetype_mro_id_array_place.index) for offset in archetype_mro_id_offsets]
            )
            self.archetype_mro_id_array_rom_indexes = archetype_mro_id_rom_indexes


class CallbackContextState:
    callback: str
    used_names: dict[str, int]
    debug_stack: list[str]
    no_eval: bool
    visitor_own_time: int

    def __init__(self, callback: str, no_eval: bool = False):
        self.callback = callback
        self.used_names = {}
        self.debug_stack = []
        self.no_eval = no_eval
        self.visitor_own_time = 0


class Context:
    project_state: ProjectContextState
    mode_state: ModeContextState
    callback_state: CallbackContextState
    statements: list[IRStmt]
    test: IRExpr
    outgoing: dict[float | None, Context]
    scope: Scope
    loop_variables: dict[str, Value]
    live: bool

    def __init__(
        self,
        project_state: ProjectContextState,
        mode_state: ModeContextState,
        callback_state: CallbackContextState,
        scope: Scope | None = None,
        live: bool = True,
        foldable_constants: dict[TempBlock, list[float | None]] | None = None,
    ):
        self.project_state = project_state
        self.mode_state = mode_state
        self.callback_state = callback_state
        self.statements = []
        self.test = IRConst(0)
        self.outgoing = {}
        self.scope = scope if scope is not None else Scope()
        self.loop_variables = {}
        self.live = live
        self.foldable_constants = foldable_constants if foldable_constants is not None else {}

    @property
    def rom(self) -> ReadOnlyMemory:
        return self.project_state.rom

    @property
    def blocks(self) -> type[Block]:
        return self.mode_state.mode.blocks

    @property
    def callback(self) -> str:
        return self.callback_state.callback

    @property
    def used_names(self) -> dict[str, int]:
        return self.callback_state.used_names

    @property
    def no_eval(self) -> bool:
        return self.callback_state.no_eval

    def check_readable(self, place: BlockPlace):
        if debug_config().unchecked_reads:
            return
        if not self.callback:
            return
        if isinstance(place.block, BlockData) and not self.is_readable(place):
            raise RuntimeError(f"Block {place.block} is not readable in {self.callback}")

    def check_writable(self, place: BlockPlace):
        if debug_config().unchecked_writes:
            return
        if not self.callback:
            return
        if isinstance(place.block, BlockData) and not self.is_writable(place):
            raise RuntimeError(f"Block {place.block} is not writable in {self.callback}")

    def is_readable(self, place: BlockPlace) -> bool:
        if debug_config().unchecked_reads:
            return True
        return self.callback and self.callback in self.blocks(place.block).readable

    def is_writable(self, place: BlockPlace) -> bool:
        if debug_config().unchecked_writes:
            return True
        return self.callback and self.callback in self.blocks(place.block).writable

    def add_statement(self, statement: IRStmt):
        if not self.live:
            return
        self.statements.append(statement)

    def add_statements(self, *statements: IRStmt):
        for statement in statements:
            self.add_statement(statement)

    def alloc(self, name: str | None = None, size: int = 1) -> BlockPlace:
        if size == 0:
            return BlockPlace(TempBlock(name or "e", 0), 0)
        name = name or ("v" if size == 1 else "a")
        num = self._get_alloc_number(name)
        if name == "v" and num < PREALLOCATED_TEMP_PLACE_COUNT:
            return preallocated_temp_block_places[num]
        return BlockPlace(TempBlock(f"{name}{num}", size), 0)

    def _get_alloc_number(self, name: str) -> int:
        if name not in self.used_names:
            self.used_names[name] = 0
        self.used_names[name] += 1
        return self.used_names[name]

    def save_alloc_state(self) -> dict[str, int]:
        return self.used_names.copy()

    def restore_alloc_state(self, state: dict[str, int]):
        self.callback_state.used_names = state.copy()

    def copy_with_scope(self, scope: Scope) -> Context:
        return Context(
            project_state=self.project_state,
            mode_state=self.mode_state,
            callback_state=self.callback_state,
            scope=scope,
            live=self.live,
            foldable_constants={**self.foldable_constants},  # We want this to persist even if the scope changes
        )

    def branch(self, condition: float | None):
        assert condition not in self.outgoing
        result = self.copy_with_scope(self.scope.copy())
        self.outgoing[condition] = result
        return result

    def branch_with_scope(self, condition: float | None, scope: Scope):
        assert condition not in self.outgoing
        result = self.copy_with_scope(scope)
        self.outgoing[condition] = result
        return result

    def new_disconnected(self):
        return self.copy_with_scope(self.scope.copy())

    def new_empty_disconnected(self):
        return self.copy_with_scope(Scope())

    def into_dead(self):
        """Create a new context for code that is unreachable, like after a return statement."""
        result = self.copy_with_scope(self.scope.copy())
        result.live = False
        return result

    def prepare_loop_header(self, to_merge: set[str]) -> Context:
        # to_merge is the set of bindings set anywhere in the loop
        # we need to invalidate them in the header if they're reference types
        # or merge them if they're value types
        # structure is self -> intermediate -> header -> body (continue -> header) | exit
        assert len(self.outgoing) == 0
        header = self.branch(None)
        for name in sorted(to_merge):
            binding = self.scope.get_binding(name)
            if not isinstance(binding, ValueBinding):
                continue
            value = binding.value
            type_ = type(value)
            if type_._is_value_type_():
                target_value = type_._from_place_(header.alloc(size=type_._size_()))
                with using_ctx(self):
                    target_value._set_(value)
                header.scope.set_value(name, target_value)
                header.loop_variables[name] = target_value
            else:
                header.scope.set_value(name, value)
                header.loop_variables[name] = value
        return header

    def branch_to_loop_header(self, header: Self):
        if not self.live:
            return
        assert len(self.outgoing) == 0
        self.outgoing[None] = header
        values = {}
        # First do a pass through and get every value
        for name, target_value in header.loop_variables.items():
            with using_ctx(self):
                if type(target_value)._is_value_type_():
                    value = self.scope.get_value(name)
                    # We make this call to _get_readonly_() to ensure that we're reading the value at this
                    # point in time specifically, since _get_readonly_ will make a copy if the value is
                    # e.g. a Num backed by a TempBlock which could be mutated.
                    values[name] = value._get_readonly_()
        # Then actually set them
        for name, target_value in header.loop_variables.items():
            with using_ctx(self):
                if type(target_value)._is_value_type_():
                    value = values[name]
                    value = type(target_value)._accept_(value)
                    target_value._set_(value)
                else:
                    value = self.scope.get_value(name)
                    if target_value is not value:
                        raise RuntimeError(
                            f"Variable '{name}' may have conflicting definitions between loop iterations"
                        )

    def map_constant(self, value: Any) -> int:
        with self.project_state.lock:
            const_mappings = self.project_state.const_mappings
            if value not in const_mappings:
                const_mappings[value] = len(const_mappings)
            return const_mappings[value]

    def map_debug_message(self, message: str) -> int:
        with self.project_state.lock:
            message_with_trace = "\n".join([*self.callback_state.debug_stack, message])
            debug_str_mappings = self.project_state.debug_str_mappings
            if message_with_trace not in debug_str_mappings:
                debug_str_mappings[message_with_trace] = len(debug_str_mappings) + 1
            return debug_str_mappings[message_with_trace]

    def get_global_base(self, value: _GlobalInfo | _GlobalPlaceholder) -> BlockPlace:
        with self.mode_state.lock:
            block = value.blocks.get(self.mode_state.mode)
            if block is None:
                raise RuntimeError(f"Global {value} is not available in '{self.mode_state.mode.name}' mode")
            if value not in self.mode_state.environment_mappings:
                if value.offset is None:
                    offset = self.mode_state.environment_offsets.get(block, 0)
                    self.mode_state.environment_mappings[value] = offset
                    self.mode_state.environment_offsets[block] = offset + value.size
                else:
                    self.mode_state.environment_mappings[value] = value.offset
            return BlockPlace(block, self.mode_state.environment_mappings[value])

    @classmethod
    def meet(cls, contexts: list[Context]) -> Context:
        if not contexts:
            raise RuntimeError("Cannot meet empty list of contexts")
        if not any(context.live for context in contexts):
            return contexts[0].into_dead()
        contexts = [context for context in contexts if context.live]
        assert not any(context.outgoing for context in contexts)
        assert all(len(context.outgoing) == 0 for context in contexts)
        target = contexts[0].copy_with_scope(Scope())
        Scope.apply_merge(target, contexts)
        for context in contexts:
            context.outgoing[None] = target
        return target

    def register_archetype(self, type_: type) -> int:
        with self.mode_state.lock:
            if type_ not in self.mode_state.archetypes:
                self.mode_state.archetypes[type_] = len(self.mode_state.archetypes)
            return self.mode_state.archetypes[type_]

    def get_archetype_mro_id_array(self, archetype_id: int) -> Sequence[int]:
        from sonolus.script.containers import ArrayPointer
        from sonolus.script.num import Num
        from sonolus.script.pointer import _deref

        self.mode_state._init_archetype_mro_info(self.rom)
        rom_index = self.mode_state.archetype_mro_id_array_rom_indexes[archetype_id]
        return ArrayPointer[int](_deref(self.blocks.EngineRom, rom_index, Num), self.blocks.EngineRom, rom_index + 1)


def ctx() -> Context | Any:  # Using Any to silence type checker warnings if it's None
    return context_var.get()


def set_ctx(value: Context | None):
    return context_var.set(value)


@contextmanager
def using_ctx(value: Context | None):
    token = context_var.set(value)
    try:
        yield
    finally:
        context_var.reset(token)


class ReadOnlyMemory:
    values: list[float]
    indexes: dict[tuple[float, ...], int]
    _lock: Lock

    def __init__(self):
        self.values = [
            float("nan"),
            float("inf"),
            float("-inf"),
        ]
        self.indexes = {}
        self._lock = Lock()

    def __getitem__(self, item: tuple[float, ...]) -> BlockPlace:
        with self._lock:
            if item not in self.indexes:
                index = len(self.values)
                self.indexes[item] = index
                self.values.extend(item)
            else:
                index = self.indexes[item]
            return BlockPlace(self.block, index)

    @property
    def block(self) -> Block:
        if ctx():
            return ctx().blocks.EngineRom
        else:
            return PlayBlock.EngineRom

    def get_value(self, index: int) -> float:
        with self._lock:
            if index < 0 or index >= len(self.values):
                raise IndexError(f"Index {index} out of bounds for ReadOnlyMemory")
            return self.values[index]


@contextmanager
def enable_debug(config: DebugConfig | None = None):
    if config is None:
        config = _full_debug_config
    token = debug_var.set(config)
    try:
        yield
    finally:
        debug_var.reset(token)


def debug_config() -> DebugConfig:
    return debug_var.get()


@dataclass
class ValueBinding:
    value: Value


@dataclass
class ConflictBinding:
    pass


@dataclass
class EmptyBinding:
    pass


Binding = ValueBinding | ConflictBinding | EmptyBinding


class Scope:
    bindings: dict[str, Binding]

    def __init__(self, bindings: dict[str, Binding] | None = None):
        self.bindings = bindings or {}

    def get_binding(self, name: str) -> Binding:
        return self.bindings.get(name, EmptyBinding())

    def set_binding(self, name: str, binding: Binding):
        self.bindings[name] = binding

    def get_value(self, name: str) -> Value | Any:
        binding = self.get_binding(name)
        match binding:
            case ValueBinding(value):
                # we don't need to call _get_() here because _set_() is never called where it could be a problem
                return value
            case ConflictBinding():
                raise RuntimeError(
                    f"Binding '{name}' has multiple conflicting definitions or may not be guaranteed to be defined"
                )
            case EmptyBinding():
                raise RuntimeError(f"Binding '{name}' is not defined")

    def set_value(self, name: str, value: Value):
        from sonolus.script.internal.impl import validate_value

        self.bindings[name] = ValueBinding(validate_value(value))

    def delete_binding(self, name: str):
        del self.bindings[name]

    def copy(self) -> Scope:
        return Scope(self.bindings.copy())

    @classmethod
    def apply_merge(cls, target: Context, incoming: list[Context]):
        if not incoming:
            return
        assert all(len(inc.outgoing) == 0 for inc in incoming)
        sources = [context.scope for context in incoming]
        keys = unique(key for source in sources for key in source.bindings)
        for key in keys:
            bindings = [source.get_binding(key) for source in sources]
            if not all(isinstance(binding, ValueBinding) for binding in bindings):
                target.scope.set_binding(key, ConflictBinding())
                continue
            values = [binding.value for binding in bindings]
            if len({id(value) for value in values}) == 1:
                target.scope.set_binding(key, ValueBinding(values[0]))
                continue
            types = {type(value) for value in values}
            if len(types) > 1:
                target.scope.set_binding(key, ConflictBinding())
                continue
            common_type: type[Value] = types.pop()
            with using_ctx(target):
                target_value = common_type._get_merge_target_(values)
            if target_value is not NotImplemented:
                for inc in incoming:
                    with using_ctx(inc):
                        target_value._set_(inc.scope.get_value(key))
                target.scope.set_value(key, target_value)
                continue
            else:
                target.scope.set_binding(key, ConflictBinding())
                continue


def iter_contexts(context: Context):
    seen = set()
    queue = [context]
    while queue:
        current = queue.pop()
        if current in seen:
            continue
        seen.add(current)
        yield current
        queue.extend(current.outgoing.values())


def context_to_cfg(context: Context) -> BasicBlock:
    blocks = {context: BasicBlock(statements=context.statements, test=context.test)}
    for current in iter_contexts(context):
        for condition, target in current.outgoing.items():
            if target not in blocks:
                blocks[target] = BasicBlock(statements=target.statements, test=target.test)
            edge = FlowEdge(src=blocks[current], dst=blocks[target], cond=condition)
            blocks[current].outgoing.add(edge)
            blocks[target].incoming.add(edge)
    result = blocks[context]
    for current in tuple(iter_contexts(context)):
        # Break cycles so memory can be cleaned without gc
        del current.outgoing
    return result


def unique[T](iterable: Iterable[T]) -> list[T]:
    result = []
    seen = set()
    for item in iterable:
        if item not in seen:
            seen.add(item)
            result.append(item)
    return result



from pydori.lib.effect import SFX_DISTANCE, Effects
from pydori.lib.layer import LAYER_JUDGE_LINE, LAYER_LANE
from pydori.lib.layout import (
    END_LANE,
    LANE_COUNT_DIM,
    START_LANE,
    layout_judge_line,
    layout_lane,
    layout_stage_left_border,
    layout_stage_right_border,
)
from pydori.lib.options import Options
from pydori.lib.particle import Particles
from pydori.lib.skin import Skin
from sonolus.script.containers import ArrayMap
from sonolus.script.globals import level_data
from sonolus.script.quad import Quad

# The duration of the lane effect particle.
LANE_EFFECT_DURATION = 0.2


@level_data
class StageData:
    lane_layouts: ArrayMap[float, Quad, LANE_COUNT_DIM]
    left_border_layout: Quad
    right_border_layout: Quad
    judge_line_layout: Quad


def init_stage_data():
    for lane in range(START_LANE, END_LANE + 1):
        StageData.lane_layouts[lane] = layout_lane(lane)
    StageData.left_border_layout = layout_stage_left_border()
    StageData.right_border_layout = layout_stage_right_border()
    StageData.judge_line_layout = layout_judge_line()


def get_lane_quad(lane: float) -> Quad:
    """Return the precomputed layout quad for a given lane."""
    return StageData.lane_layouts[lane]


def draw_stage():
    """Draw the stage, including lanes, borders, and the judgment line."""
    for lane_quad in StageData.lane_layouts.values():
        Skin.lane.draw(lane_quad, z=LAYER_LANE)
    Skin.stage_left_border.draw(StageData.left_border_layout, z=LAYER_LANE)
    Skin.stage_right_border.draw(StageData.right_border_layout, z=LAYER_LANE)
    Skin.judgment_line.draw(StageData.judge_line_layout, z=LAYER_JUDGE_LINE)


def play_lane_sfx():
    if not Options.sfx_enabled:
        return
    Effects.stage.play(SFX_DISTANCE)


def schedule_lane_sfx(target_time: float):
    if not Options.sfx_enabled:
        return
    Effects.stage.schedule(target_time, SFX_DISTANCE)


def play_lane_particle(lane: float):
    if not Options.lane_effect_enabled:
        return
    Particles.lane.spawn(
        get_lane_quad(lane),
        duration=LANE_EFFECT_DURATION,
    )



import sys
from pathlib import Path

from sonolus.script.project import Project

sys.path.insert(0, str(Path(__file__).parent.parent.parent / "test_projects"))

import pydori.project

pydori_project: Project = pydori.project.project



from sonolus.script.archetype import StandardArchetypeName, StandardImport, WatchArchetype, imported


class WatchBpmChange(WatchArchetype):
    name = StandardArchetypeName.BPM_CHANGE

    beat: StandardImport.BEAT = imported()
    bpm: StandardImport.BPM = imported()


class WatchTimescaleChange(WatchArchetype):
    name = StandardArchetypeName.TIMESCALE_CHANGE

    beat: StandardImport.BEAT = imported()
    timescale: StandardImport.TIMESCALE = imported()



from enum import StrEnum


class StandardText(StrEnum):
    """Standard text constants."""

    POST = "#POST"
    PLAYLIST = "#PLAYLIST"
    LEVEL = "#LEVEL"
    SKIN = "#SKIN"
    BACKGROUND = "#BACKGROUND"
    EFFECT = "#EFFECT"
    PARTICLE = "#PARTICLE"
    ENGINE = "#ENGINE"
    REPLAY = "#REPLAY"
    ROOM = "#ROOM"
    GRADE = "#GRADE"
    ARCADE_SCORE = "#ARCADE_SCORE"
    ACCURACY_SCORE = "#ACCURACY_SCORE"
    COMBO = "#COMBO"
    PERFECT = "#PERFECT"
    GREAT = "#GREAT"
    GOOD = "#GOOD"
    MISS = "#MISS"
    KEYWORDS = "#KEYWORDS"
    NAME = "#NAME"
    RATING = "#RATING"
    RATING_MINIMUM = "#RATING_MINIMUM"
    RATING_MAXIMUM = "#RATING_MAXIMUM"
    TITLE = "#TITLE"
    SUBTITLE = "#SUBTITLE"
    ARTISTS = "#ARTISTS"
    TIME = "#TIME"
    AUTHOR = "#AUTHOR"
    GENRE = "#GENRE"
    TYPE = "#TYPE"
    CATEGORY = "#CATEGORY"
    LANGUAGE = "#LANGUAGE"
    DIFFICULTY = "#DIFFICULTY"
    VERSION = "#VERSION"
    LENGTH = "#LENGTH"
    LENGTH_MINIMUM = "#LENGTH_MINIMUM"
    LENGTH_MAXIMUM = "#LENGTH_MAXIMUM"
    ADDITIONAL_INFORMATION = "#ADDITIONAL_INFORMATION"
    TIMEZONE = "#TIMEZONE"
    REGION = "#REGION"
    CONTENT = "#CONTENT"
    COMMENT = "#COMMENT"
    REVIEW = "#REVIEW"
    REPLY = "#REPLY"
    MESSAGE = "#MESSAGE"
    ROLE = "#ROLE"
    PERMISSION = "#PERMISSION"
    USER = "#USER"
    SPEED = "#SPEED"
    MIRROR = "#MIRROR"
    RANDOM = "#RANDOM"
    HIDDEN = "#HIDDEN"
    JUDGMENT_STRICT = "#JUDGMENT_STRICT"
    JUDGMENT_LOOSE = "#JUDGMENT_LOOSE"
    EFFECT_AUTO = "#EFFECT_AUTO"
    STAGE = "#STAGE"
    STAGE_POSITION = "#STAGE_POSITION"
    STAGE_SIZE = "#STAGE_SIZE"
    STAGE_ROTATION = "#STAGE_ROTATION"
    STAGE_DIRECTION = "#STAGE_DIRECTION"
    STAGE_ALPHA = "#STAGE_ALPHA"
    STAGE_ANIMATION = "#STAGE_ANIMATION"
    STAGE_TILT = "#STAGE_TILT"
    STAGE_COVER_VERTICAL = "#STAGE_COVER_VERTICAL"
    STAGE_COVER_HORIZONTAL = "#STAGE_COVER_HORIZONTAL"
    STAGE_COVER_ALPHA = "#STAGE_COVER_ALPHA"
    STAGE_ASPECTRATIO_LOCK = "#STAGE_ASPECTRATIO_LOCK"
    STAGE_EFFECT = "#STAGE_EFFECT"
    STAGE_EFFECT_POSITION = "#STAGE_EFFECT_POSITION"
    STAGE_EFFECT_SIZE = "#STAGE_EFFECT_SIZE"
    STAGE_EFFECT_ROTATION = "#STAGE_EFFECT_ROTATION"
    STAGE_EFFECT_DIRECTION = "#STAGE_EFFECT_DIRECTION"
    STAGE_EFFECT_ALPHA = "#STAGE_EFFECT_ALPHA"
    LANE = "#LANE"
    LANE_POSITION = "#LANE_POSITION"
    LANE_SIZE = "#LANE_SIZE"
    LANE_ROTATION = "#LANE_ROTATION"
    LANE_DIRECTION = "#LANE_DIRECTION"
    LANE_ALPHA = "#LANE_ALPHA"
    LANE_ANIMATION = "#LANE_ANIMATION"
    LANE_EFFECT = "#LANE_EFFECT"
    LANE_EFFECT_POSITION = "#LANE_EFFECT_POSITION"
    LANE_EFFECT_SIZE = "#LANE_EFFECT_SIZE"
    LANE_EFFECT_ROTATION = "#LANE_EFFECT_ROTATION"
    LANE_EFFECT_DIRECTION = "#LANE_EFFECT_DIRECTION"
    LANE_EFFECT_ALPHA = "#LANE_EFFECT_ALPHA"
    JUDGELINE = "#JUDGELINE"
    JUDGELINE_POSITION = "#JUDGELINE_POSITION"
    JUDGELINE_SIZE = "#JUDGELINE_SIZE"
    JUDGELINE_ROTATION = "#JUDGELINE_ROTATION"
    JUDGELINE_DIRECTION = "#JUDGELINE_DIRECTION"
    JUDGELINE_ALPHA = "#JUDGELINE_ALPHA"
    JUDGELINE_ANIMATION = "#JUDGELINE_ANIMATION"
    JUDGELINE_EFFECT = "#JUDGELINE_EFFECT"
    JUDGELINE_EFFECT_POSITION = "#JUDGELINE_EFFECT_POSITION"
    JUDGELINE_EFFECT_SIZE = "#JUDGELINE_EFFECT_SIZE"
    JUDGELINE_EFFECT_ROTATION = "#JUDGELINE_EFFECT_ROTATION"
    JUDGELINE_EFFECT_DIRECTION = "#JUDGELINE_EFFECT_DIRECTION"
    JUDGELINE_EFFECT_ALPHA = "#JUDGELINE_EFFECT_ALPHA"
    SLOT = "#SLOT"
    SLOT_POSITION = "#SLOT_POSITION"
    SLOT_SIZE = "#SLOT_SIZE"
    SLOT_ROTATION = "#SLOT_ROTATION"
    SLOT_DIRECTION = "#SLOT_DIRECTION"
    SLOT_ALPHA = "#SLOT_ALPHA"
    SLOT_ANIMATION = "#SLOT_ANIMATION"
    SLOT_EFFECT = "#SLOT_EFFECT"
    SLOT_EFFECT_POSITION = "#SLOT_EFFECT_POSITION"
    SLOT_EFFECT_SIZE = "#SLOT_EFFECT_SIZE"
    SLOT_EFFECT_ROTATION = "#SLOT_EFFECT_ROTATION"
    SLOT_EFFECT_DIRECTION = "#SLOT_EFFECT_DIRECTION"
    SLOT_EFFECT_ALPHA = "#SLOT_EFFECT_ALPHA"
    NOTE = "#NOTE"
    NOTE_SPEED = "#NOTE_SPEED"
    NOTE_SPEED_RANDOM = "#NOTE_SPEED_RANDOM"
    NOTE_POSITION = "#NOTE_POSITION"
    NOTE_SIZE = "#NOTE_SIZE"
    NOTE_ROTATION = "#NOTE_ROTATION"
    NOTE_DIRECTION = "#NOTE_DIRECTION"
    NOTE_COLOR = "#NOTE_COLOR"
    NOTE_ALPHA = "#NOTE_ALPHA"
    NOTE_ANIMATION = "#NOTE_ANIMATION"
    NOTE_EFFECT = "#NOTE_EFFECT"
    NOTE_EFFECT_POSITION = "#NOTE_EFFECT_POSITION"
    NOTE_EFFECT_SIZE = "#NOTE_EFFECT_SIZE"
    NOTE_EFFECT_ROTATION = "#NOTE_EFFECT_ROTATION"
    NOTE_EFFECT_DIRECTION = "#NOTE_EFFECT_DIRECTION"
    NOTE_EFFECT_COLOR = "#NOTE_EFFECT_COLOR"
    NOTE_EFFECT_ALPHA = "#NOTE_EFFECT_ALPHA"
    MARKER = "#MARKER"
    MARKER_POSITION = "#MARKER_POSITION"
    MARKER_SIZE = "#MARKER_SIZE"
    MARKER_ROTATION = "#MARKER_ROTATION"
    MARKER_DIRECTION = "#MARKER_DIRECTION"
    MARKER_COLOR = "#MARKER_COLOR"
    MARKER_ALPHA = "#MARKER_ALPHA"
    MARKER_ANIMATION = "#MARKER_ANIMATION"
    CONNECTOR = "#CONNECTOR"
    CONNECTOR_POSITION = "#CONNECTOR_POSITION"
    CONNECTOR_SIZE = "#CONNECTOR_SIZE"
    CONNECTOR_ROTATION = "#CONNECTOR_ROTATION"
    CONNECTOR_DIRECTION = "#CONNECTOR_DIRECTION"
    CONNECTOR_COLOR = "#CONNECTOR_COLOR"
    CONNECTOR_ALPHA = "#CONNECTOR_ALPHA"
    CONNECTOR_ANIMATION = "#CONNECTOR_ANIMATION"
    SIMLINE = "#SIMLINE"
    SIMLINE_POSITION = "#SIMLINE_POSITION"
    SIMLINE_SIZE = "#SIMLINE_SIZE"
    SIMLINE_ROTATION = "#SIMLINE_ROTATION"
    SIMLINE_DIRECTION = "#SIMLINE_DIRECTION"
    SIMLINE_COLOR = "#SIMLINE_COLOR"
    SIMLINE_ALPHA = "#SIMLINE_ALPHA"
    SIMLINE_ANIMATION = "#SIMLINE_ANIMATION"
    PREVIEW_SCALE_VERTICAL = "#PREVIEW_SCALE_VERTICAL"
    PREVIEW_SCALE_HORIZONTAL = "#PREVIEW_SCALE_HORIZONTAL"
    PREVIEW_TIME = "#PREVIEW_TIME"
    PREVIEW_SCORE = "#PREVIEW_SCORE"
    PREVIEW_BPM = "#PREVIEW_BPM"
    PREVIEW_TIMESCALE = "#PREVIEW_TIMESCALE"
    PREVIEW_BEAT = "#PREVIEW_BEAT"
    PREVIEW_MEASURE = "#PREVIEW_MEASURE"
    PREVIEW_COMBO = "#PREVIEW_COMBO"
    NONE = "#NONE"
    ANY = "#ANY"
    ALL = "#ALL"
    OTHERS = "#OTHERS"
    SHORT = "#SHORT"
    LONG = "#LONG"
    HIGH = "#HIGH"
    MID = "#MID"
    LOW = "#LOW"
    SMALL = "#SMALL"
    MEDIUM = "#MEDIUM"
    LARGE = "#LARGE"
    LEFT = "#LEFT"
    RIGHT = "#RIGHT"
    UP = "#UP"
    DOWN = "#DOWN"
    FRONT = "#FRONT"
    BACK = "#BACK"
    CENTER = "#CENTER"
    TOP = "#TOP"
    BOTTOM = "#BOTTOM"
    TOP_LEFT = "#TOP_LEFT"
    TOP_CENTER = "#TOP_CENTER"
    TOP_RIGHT = "#TOP_RIGHT"
    CENTER_LEFT = "#CENTER_LEFT"
    CENTER_RIGHT = "#CENTER_RIGHT"
    BOTTOM_LEFT = "#BOTTOM_LEFT"
    BOTTOM_CENTER = "#BOTTOM_CENTER"
    BOTTOM_RIGHT = "#BOTTOM_RIGHT"
    CLOCKWISE = "#CLOCKWISE"
    COUNTERCLOCKWISE = "#COUNTERCLOCKWISE"
    FORWARD = "#FORWARD"
    BACKWARD = "#BACKWARD"
    DEFAULT = "#DEFAULT"
    NEUTRAL = "#NEUTRAL"
    RED = "#RED"
    GREEN = "#GREEN"
    BLUE = "#BLUE"
    YELLOW = "#YELLOW"
    PURPLE = "#PURPLE"
    CYAN = "#CYAN"
    SIMPLE = "#SIMPLE"
    EASY = "#EASY"
    NORMAL = "#NORMAL"
    HARD = "#HARD"
    EXPERT = "#EXPERT"
    MASTER = "#MASTER"
    PRO = "#PRO"
    TECHNICAL = "#TECHNICAL"
    SPECIAL = "#SPECIAL"
    KEYWORDS_PLACEHOLDER = "#KEYWORDS_PLACEHOLDER"
    NAME_PLACEHOLDER = "#NAME_PLACEHOLDER"
    RATING_PLACEHOLDER = "#RATING_PLACEHOLDER"
    RATING_MINIMUM_PLACEHOLDER = "#RATING_MINIMUM_PLACEHOLDER"
    RATING_MAXIMUM_PLACEHOLDER = "#RATING_MAXIMUM_PLACEHOLDER"
    TITLE_PLACEHOLDER = "#TITLE_PLACEHOLDER"
    SUBTITLE_PLACEHOLDER = "#SUBTITLE_PLACEHOLDER"
    ARTISTS_PLACEHOLDER = "#ARTISTS_PLACEHOLDER"
    TIME_PLACEHOLDER = "#TIME_PLACEHOLDER"
    AUTHOR_PLACEHOLDER = "#AUTHOR_PLACEHOLDER"
    GENRE_PLACEHOLDER = "#GENRE_PLACEHOLDER"
    TYPE_PLACEHOLDER = "#TYPE_PLACEHOLDER"
    CATEGORY_PLACEHOLDER = "#CATEGORY_PLACEHOLDER"
    LANGUAGE_PLACEHOLDER = "#LANGUAGE_PLACEHOLDER"
    DIFFICULTY_PLACEHOLDER = "#DIFFICULTY_PLACEHOLDER"
    LENGTH_PLACEHOLDER = "#LENGTH_PLACEHOLDER"
    LENGTH_MINIMUM_PLACEHOLDER = "#LENGTH_MINIMUM_PLACEHOLDER"
    LENGTH_MAXIMUM_PLACEHOLDER = "#LENGTH_MAXIMUM_PLACEHOLDER"
    ADDITIONAL_INFORMATION_PLACEHOLDER = "#ADDITIONAL_INFORMATION_PLACEHOLDER"
    TIMEZONE_PLACEHOLDER = "#TIMEZONE_PLACEHOLDER"
    REGION_PLACEHOLDER = "#REGION_PLACEHOLDER"
    CONTENT_PLACEHOLDER = "#CONTENT_PLACEHOLDER"
    COMMENT_PLACEHOLDER = "#COMMENT_PLACEHOLDER"
    REVIEW_PLACEHOLDER = "#REVIEW_PLACEHOLDER"
    REPLY_PLACEHOLDER = "#REPLY_PLACEHOLDER"
    MESSAGE_PLACEHOLDER = "#MESSAGE_PLACEHOLDER"
    ROLE_PLACEHOLDER = "#ROLE_PLACEHOLDER"
    PERMISSION_PLACEHOLDER = "#PERMISSION_PLACEHOLDER"
    USER_PLACEHOLDER = "#USER_PLACEHOLDER"
    PERCENTAGE_UNIT = "#PERCENTAGE_UNIT"
    SECOND_UNIT = "#SECOND_UNIT"
    MILLISECOND_UNIT = "#MILLISECOND_UNIT"
    TAP = "#TAP"
    TAP_HOLD = "#TAP_HOLD"
    TAP_RELEASE = "#TAP_RELEASE"
    TAP_FLICK = "#TAP_FLICK"
    TAP_SLIDE = "#TAP_SLIDE"
    HOLD = "#HOLD"
    HOLD_SLIDE = "#HOLD_SLIDE"
    HOLD_FOLLOW = "#HOLD_FOLLOW"
    RELEASE = "#RELEASE"
    FLICK = "#FLICK"
    SLIDE = "#SLIDE"
    SLIDE_FLICK = "#SLIDE_FLICK"
    AVOID = "#AVOID"
    JIGGLE = "#JIGGLE"
    NEWEST = "#NEWEST"
    OLDEST = "#OLDEST"
    RECOMMENDED = "#RECOMMENDED"
    POPULAR = "#POPULAR"
    FEATURED = "#FEATURED"
    COMPETITIVE = "#COMPETITIVE"
    HOLIDAY = "#HOLIDAY"
    LIMITED = "#LIMITED"
    FOLLOWING = "#FOLLOWING"
    ANNOUNCEMENT = "#ANNOUNCEMENT"
    INFORMATION = "#INFORMATION"
    HELP = "#HELP"
    MAINTENANCE = "#MAINTENANCE"
    EVENT = "#EVENT"
    UPDATE = "#UPDATE"
    SEARCH = "#SEARCH"
    ADVANCED = "#ADVANCED"
    RELATED = "#RELATED"
    SAME_AUTHOR = "#SAME_AUTHOR"
    SAME_ARTISTS = "#SAME_ARTISTS"
    SAME_RATING = "#SAME_RATING"
    SAME_CATEGORY = "#SAME_CATEGORY"
    SAME_DIFFICULTY = "#SAME_DIFFICULTY"
    SAME_GENRE = "#SAME_GENRE"
    SAME_VERSION = "#SAME_VERSION"
    OTHER_AUTHORS = "#OTHER_AUTHORS"
    OTHER_ARTISTS = "#OTHER_ARTISTS"
    OTHER_RATINGS = "#OTHER_RATINGS"
    OTHER_CATEGORIES = "#OTHER_CATEGORIES"
    OTHER_DIFFICULTIES = "#OTHER_DIFFICULTIES"
    OTHER_GENRES = "#OTHER_GENRES"
    OTHER_VERSIONS = "#OTHER_VERSIONS"
    DRAFT = "#DRAFT"
    REVIEWING = "#REVIEWING"
    PUBLISHED = "#PUBLISHED"
    VERIFYING = "#VERIFYING"
    VERIFIED = "#VERIFIED"
    REMOVED = "#REMOVED"
    PUBLIC = "#PUBLIC"
    PRIVATE = "#PRIVATE"
    LOCKED = "#LOCKED"
    POP = "#POP"
    ROCK = "#ROCK"
    HIPHOP = "#HIPHOP"
    COUNTRY = "#COUNTRY"
    ELECTRONIC = "#ELECTRONIC"
    METAL = "#METAL"
    CLASSICAL = "#CLASSICAL"
    FOLK = "#FOLK"
    INDIE = "#INDIE"
    ANIME = "#ANIME"
    VOCALOID = "#VOCALOID"
    REMIX = "#REMIX"
    INSTRUMENTAL = "#INSTRUMENTAL"
    SHORT_VERSION = "#SHORT_VERSION"
    LONG_VERSION = "#LONG_VERSION"
    LIVE_VERSION = "#LIVE_VERSION"
    REPORT = "#REPORT"
    REASON = "#REASON"
    ILLEGAL_ACTIVITIES = "#ILLEGAL_ACTIVITIES"
    CHEATING = "#CHEATING"
    AFK = "#AFK"
    SPAMMING = "#SPAMMING"
    VERBAL_ABUSE = "#VERBAL_ABUSE"
    INAPPROPRIATE_LANGUAGE = "#INAPPROPRIATE_LANGUAGE"
    NEGATIVE_ATTITUDE = "#NEGATIVE_ATTITUDE"
    DNF = "#DNF"
    SUGGESTIONS = "#SUGGESTIONS"
    SUGGESTIONS_PER_PLAYER = "#SUGGESTIONS_PER_PLAYER"
    MATCH_SCORING = "#MATCH_SCORING"
    MATCH_TIEBREAKER = "#MATCH_TIEBREAKER"
    MATCH_COUNT = "#MATCH_COUNT"
    MATCH_LIMIT = "#MATCH_LIMIT"
    ROUND_SCORING = "#ROUND_SCORING"
    ROUND_TIEBREAKER = "#ROUND_TIEBREAKER"
    ROUND_COUNT = "#ROUND_COUNT"
    ROUND_LIMIT = "#ROUND_LIMIT"
    TEAM_SCORING = "#TEAM_SCORING"
    TEAM_TIEBREAKER = "#TEAM_TIEBREAKER"
    TEAM_COUNT = "#TEAM_COUNT"
    TEAM_LIMIT = "#TEAM_LIMIT"
    QUALIFIED = "#QUALIFIED"
    DISQUALIFIED = "#DISQUALIFIED"
    RANKING = "#RANKING"
    SCORE = "#SCORE"
    PLAYER = "#PLAYER"
    SPECTATOR = "#SPECTATOR"
    REFEREE = "#REFEREE"
    ELIMINATED = "#ELIMINATED"
    FINALIST = "#FINALIST"
    FINISHED = "#FINISHED"
    WINNER = "#WINNER"
    GOLD_MEDAL = "#GOLD_MEDAL"
    SILVER_MEDAL = "#SILVER_MEDAL"
    BRONZE_MEDAL = "#BRONZE_MEDAL"
    TEAM_1 = "#TEAM_1"
    TEAM_2 = "#TEAM_2"
    TEAM_3 = "#TEAM_3"
    TEAM_4 = "#TEAM_4"
    TEAM_5 = "#TEAM_5"
    TEAM_6 = "#TEAM_6"
    TEAM_7 = "#TEAM_7"
    TEAM_8 = "#TEAM_8"
    TEAM_RED = "#TEAM_RED"
    TEAM_GREEN = "#TEAM_GREEN"
    TEAM_BLUE = "#TEAM_BLUE"
    TEAM_YELLOW = "#TEAM_YELLOW"
    TEAM_PURPLE = "#TEAM_PURPLE"
    TEAM_CYAN = "#TEAM_CYAN"
    TEAM_WHITE = "#TEAM_WHITE"
    TEAM_BLACK = "#TEAM_BLACK"
    UPLOAD = "#UPLOAD"
    SUBMIT = "#SUBMIT"
    EDIT = "#EDIT"
    LIKE = "#LIKE"
    LIKED = "#LIKED"
    DISLIKE = "#DISLIKE"
    DISLIKED = "#DISLIKED"
    BOOKMARK = "#BOOKMARK"
    BOOKMARKED = "#BOOKMARKED"
    DELETE = "#DELETE"
    RESTORE = "#RESTORE"
    CONFIRM = "#CONFIRM"
    CANCEL = "#CANCEL"
    INCREASE = "#INCREASE"
    DECREASE = "#DECREASE"
    UPVOTE = "#UPVOTE"
    UPVOTED = "#UPVOTED"
    DOWNVOTE = "#DOWNVOTE"
    DOWNVOTED = "#DOWNVOTED"
    AGREE = "#AGREE"
    AGREED = "#AGREED"
    DISAGREE = "#DISAGREE"
    DISAGREED = "#DISAGREED"
    LOCK = "#LOCK"
    UNLOCK = "#UNLOCK"
    PIN = "#PIN"
    UNPIN = "#UNPIN"
    FOLLOW = "#FOLLOW"
    UNFOLLOW = "#UNFOLLOW"
    SUBSCRIBE = "#SUBSCRIBE"
    UNSUBSCRIBE = "#UNSUBSCRIBE"
    PUBLISH = "#PUBLISH"
    UNPUBLISH = "#UNPUBLISH"
    SHOW = "#SHOW"
    HIDE = "#HIDE"
    ALLOW = "#ALLOW"
    DISALLOW = "#DISALLOW"
    APPROVE = "#APPROVE"
    DENY = "#DENY"
    ACCEPT = "#ACCEPT"
    REJECT = "#REJECT"
    STAR = "#STAR"
    STARRED = "#STARRED"
